from fastapi import FastAPI, Request, status, Query, File, UploadFile, Form
from fastapi.responses import HTMLResponse, RedirectResponse, Response, FileResponse
#from typing import List, Optional
from fastapi.templating import Jinja2Templates
from fastapi.staticfiles import StaticFiles

from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware
import typing
import time
import subprocess
import shutil
import os
import re
from pathlib import Path
import pickle
# import dotenv
# from dotenv import load_dotenv
#import wand # evtl. imagemagick umstellen?
# from wand.image import Image as wi
# from wand.drawing import Drawing as wdraw
# from wand.color import Color
import math
import requests
from PIL import Image, ImageDraw, ImageFont
import socket

#eigene
from logdef import *
# logging.info(f"backend.app.main -->  start")

# collaps all mac command+K+0

########################
#
#  pseudo - Flash
#
#  

def flash(request: Request, message: typing.Any, category: str = "primary") -> None:
    if "_messages" not in request.session:
       request.session["_messages"] = []
    request.session["_messages"].append({"message": message, "category": category})

def get_flashed_messages(request: Request):
    return request.session.pop("_messages") if "_messages" in request.session else []

middleware = [ 
Middleware(SessionMiddleware, secret_key='super-secret') 
] 
app = FastAPI(middleware=middleware)

app.mount("/static", StaticFiles(directory="static"), name="static")
templates = Jinja2Templates(directory="templates")
templates.env.globals['get_flashed_messages'] = get_flashed_messages


#####################
# parameter save load
# tranformation
def load_parameters():
    try:
        calib_result_pickle = pickle.load(open("static/_data/_trans.pkl", "rb"))
        U = calib_result_pickle["U"]
        u = calib_result_pickle["u"]
        m = calib_result_pickle["m"]
        dpi = calib_result_pickle["dpi"]
        ret = True
    except:
        #print('Fehler in load_parameters _ Kamerakalibrierung fehlt!')
        ret = False
        U = 280
        u = 200
        m = 50
        dpi = 300
    return ret, U,u,m,dpi

def save_parameters(U,u,m,dpi):
    # 4 punkte für transformation speichern
    #print('save Parameter pts:', pts)
    trans_result_pickle = {}
    trans_result_pickle["U"] = int(U)
    trans_result_pickle["u"] = int(u)
    trans_result_pickle["m"] = int(m)
    trans_result_pickle["dpi"] = int(dpi)
    pickle.dump(trans_result_pickle, open("static/_data/_trans.pkl", "wb" ))
    return


# drucken
def load_print_para(file = "static/_data/_print_para.pkl"):
    try:
        print_para_pickle = pickle.load(open(file, "rb"))
        papier_format = print_para_pickle["papier_format"]
        papier_richtung = print_para_pickle["papier_richtung"]
        papier_rand = print_para_pickle["papier_rand"]
        zentriert = print_para_pickle["zentriert"]
        not_print = print_para_pickle["not_print"]
        show_img_info=print_para_pickle["show_img_info"]
        selcted_printer=print_para_pickle["selcted_printer"]
        spiegeln=print_para_pickle["spiegeln"]
        project_name=print_para_pickle["project_name"]
        project_besch=print_para_pickle["project_besch"]
        ret = True
    except:
        ret = False
        papier_format = 'A4'
        papier_richtung = 'quer'
        papier_rand = 15
        zentriert = 0
        not_print = []
        show_img_info = 'on'
        selcted_printer = None
        spiegeln = 'off'
        project_name = ''
        project_besch = ''

    return ret, papier_format,papier_richtung,papier_rand, zentriert, \
            not_print, show_img_info, selcted_printer, spiegeln, project_name,project_besch

def save_print_para(papier_format,papier_richtung,papier_rand,zentriert, \
                    not_print,show_img_info,selcted_printer,spiegeln,project_name,project_besch):
    trans_result_pickle = {}
    trans_result_pickle["papier_format"] = papier_format
    trans_result_pickle["papier_richtung"] = papier_richtung
    trans_result_pickle["papier_rand"] = papier_rand
    trans_result_pickle["zentriert"] = zentriert
    trans_result_pickle["not_print"] = not_print
    trans_result_pickle["show_img_info"] = show_img_info
    trans_result_pickle["selcted_printer"] = selcted_printer
    trans_result_pickle["spiegeln"] = spiegeln
    trans_result_pickle["project_name"] = project_name
    trans_result_pickle["project_besch"] = project_besch
    pickle.dump(trans_result_pickle, open("static/_data/_print_para.pkl", "wb" ))
    return

# drucker-liste -Todo:  -> doch besser immer direkt cups abfragen
def load_printer_list():
    #params = ['lp', '-h', 'cups', '-d', 'brother_pas','static/_print_jobs/Seite_0.pdf']
    params = ['lpstat', '-h', 'cups', '-v']
    printers = []
    try:
        ret = subprocess.check_output(params)
        ret_split = re.split(" ",str(ret))
        for i in range(len(ret_split)-1):
            if ret_split[i] == 'for':
                name =f"{ret_split[i+1]}"
                printers.append(name[:-1])
        ret = True
        #logging.info(f"druckerliste abfrage cups .backend.app.main -->  {ret}")
    except:
        ret = False
        logging.info(f"druckerliste abfrage cups .backend.app.main -->  ERROR")
    return ret, printers


####################
# pixel zu mm
def pix2mm(w,h,dpi_w, dpi_h):
    inch2mm = 25.4
    wmm = round(w*inch2mm/dpi_w)
    hmm = round(h*inch2mm/dpi_h)
    return [wmm,hmm]

def mm2pix(w,h,dpi):
    inch2mm = 25.4
    wpix = round(w*dpi/inch2mm)
    hpix = round(h*dpi/inch2mm)
    return [wpix,hpix]


########################
# routes
@app.get("/", response_class=HTMLResponse)
def read_root(request: Request):
    return RedirectResponse("index")

@app.get("/index", response_class=HTMLResponse)
def index(request: Request, mess_danger: str = None):
    cache_para = int(time.time()) # um browser zum reload der bilder zu 'überreden'
    ret,U,u,m,dpi = load_parameters()
    if not os.path.isfile('static/_data/ID_0.png'):
        flash(request, f"Bild hochladen!  - Datei im gestrichelten Rahmen ablegen oder in das Feld tippen um upload zu starten. Unterstützte Formate: ['.png','.jpg','.jpeg','.bmp','.tiff','.psd']","flush_danger")
    elif not os.path.isfile('static/_data/ID_2.jpg'):
        flash(request, f"Bild noch nicht transformiert!  -  [Originalmasse übernehmem] oder Parameter einstellen und [Transformieren]  anwählen", "flush_danger")

    return templates.TemplateResponse("index.html", {"request": request,
                                                     "cache_para": cache_para,
                                                     "U": U,
                                                     "u":u,
                                                     "m":m,
                                                     "dpi":dpi})

@app.get("/upload", response_class=HTMLResponse)
def upload(request: Request):
    return templates.TemplateResponse("upload.html", {"request": request})

@app.get("/system", response_class=HTMLResponse)
def system(request: Request):
    #infos:
    client_ip = request.client.host
    referrer = re.split("/", str(request.url))
    host_name = referrer[2]
    host_name2 = os.uname()[1]
    host_ip = socket.gethostbyname(host_name)
    host_ip_ct = socket.gethostbyname('backend')

    try:
        ip_ext = requests.get('http://ip.pasch.ch').content.decode()
    except:
        ip_ext = "None"

    return templates.TemplateResponse("system.html", {"request": request,
                                                    "client_ip": client_ip,
                                                    "host_name": host_name,
                                                    "host_name2": host_name2,
                                                    "host_ip": host_ip,
                                                    "host_ip_ct": host_ip_ct,
                                                    "ip_ext": ip_ext})

@app.get("/help", response_class=HTMLResponse)
def help(request: Request):
    client_ip = request.client.host
    # Einstellungen
    referrer = re.split("/", str(request.url))
    host_name = referrer[2]

    return templates.TemplateResponse("help.html", {"request": request,
                                                    "client_ip": client_ip,
                                                    "host_name": host_name})

@app.get("/work", response_class=HTMLResponse)
def work(request: Request):
    return templates.TemplateResponse("work.html",{"request": request})

###############
# projekte speichern
@app.get("/save_project", response_class=HTMLResponse)
def save_project(request: Request,
              project_name: str = '',
              project_besch: str = ''):
    
    # abbrechen wenn name fehlt:
    if project_name == '':
        flash(request, f"Projekt_name fehlt", "flush_danger")
        return RedirectResponse("print")


    # gespeicherte druckparameter lesen
    ret, papier_format,papier_richtung,papier_rand, zentriert, \
    not_print,show_img_info,selcted_printer,spiegeln,old_project_name,old_project_besch = load_print_para()
    # druckparameter aktualisieren
    save_print_para(papier_format=papier_format,
                    papier_richtung=papier_richtung,
                    papier_rand=papier_rand,
                    zentriert=zentriert,
                    not_print=not_print,
                    show_img_info=show_img_info,
                    selcted_printer=selcted_printer,
                    spiegeln=spiegeln,
                    project_name=project_name,
                    project_besch=project_besch)



    path = 'static/_data/_projects'
    project_folder = os.path.join(path, project_name)

    # ordner für neues projekt vorbereiten
    # Folder _projects erstellen falls nicht vorhanden
    if not os.path.exists(path):
        try: 
            os.mkdir(path)
            flash(request, "Ordner _projects erstellt", "flush_success")
        except OSError as error: 
            flash(request, f"Ordner _projects  konnte nicht erstellt werden - {error}", "flush_danger")
    # Projektordner erstellen oder leeren falls schon exist
    if not os.path.exists(project_folder):
        try: 
            os.mkdir(project_folder)
            #flash(request, f"Projekt-Ordner << {project_name} >> erstellt", "flush_success")
        except OSError as error: 
            flash(request, f"Projekt-Ordner konnte nicht erstellt werden - {error}", "flush_danger")
    else:
        for f in os.listdir(project_folder):
            if os.path.isfile(f):
                os.remove(os.path.join(project_folder, f))            

    # daten (nur Configs und als png gespeichertes "original" Bild) zu projekt-folder kopieren
    # -> doch besser gleich fast alles zu kopieren
    suffixes = (".pkl", ".png",".pdf",".jpg")
    for f in os.listdir('static/_data'):
        if f.endswith(suffixes):
            old_file = os.path.join('static/_data', f)
            new_file = os.path.join(project_folder, f)
            shutil.copyfile(old_file,new_file)
    
    suffixes = (".pdf")
    for f in os.listdir('static/_print_jobs'):
        if f.endswith(suffixes):
            old_file = os.path.join('static/_print_jobs', f)
            new_file = os.path.join(project_folder, f)
            shutil.copyfile(old_file,new_file)
    
    flash(request, f"Projekt \"{project_name}\" gespeicher - Menüpunkt < Projekte > ", "flush_success")
    
    return RedirectResponse("print")


####################
# projekte verwalten
@app.get("/print_queue", response_class=HTMLResponse)
def print_queue(request: Request,
              del_job: str = ''):
    
    # printernamen abfragen lpstat -h cups -v
    ret, printers = load_printer_list()
    if ret:
        flash(request, f"CUPS-Druckerliste erfolgreich abgefragt", "flush_success")
    else:
        flash(request, "CUPS-Druckerliste fehlgeschlagen", "flush_danger")

    # job löschen?
    if del_job != '':
        params = ['cancel','-h','cups', f'{del_job}']
        try:
            ret = subprocess.check_output(params)

            os.remove(f'static/_open_print_jobs/{del_job}.png')


            flash(request, f"Druckaufrtag ID: {del_job} gelöscht!", "flush_success")
        except:
            flash(request, f"Druckaufrtag ID: {del_job} nicht gelöscht! {ret}", "flush_danger")


    # offene aufträge abfragen je printer lpstat -h cups -W not-completed -o -p brother_pas  ---- lpstat -W completed -o -p
    def load_print_jobs(printer='CUPS-PDF'):
        params = ['lpstat', '-h', 'cups','-W', 'not-completed', '-o']
        jobs = []
        jobs_id_only = []
        try:
            ret = str(subprocess.check_output(params))
            #logging.info(f"druckerliste ret_split abfrage cups .backend.app.main -->  {ret}")
            ret_split = re.split(r"\\n", ret[2:])
            #logging.info(f"druckerliste ret_split abfrage cups .backend.app.main -->  {ret_split}")
            for i in ret_split:
                job = i.split()
                job_id = job[0]
                if str(job[0]).find(printer) != -1:
                    job_id = job[0].replace(f'{printer}-','')
                    jobs_id_only.append(job_id)
                    job_start_datum = "-".join(job[3:8])
                    jobs.append([job_id, printer,job_start_datum])
            #logging.info(f"Druck jobs auf {printer} -->  {jobs}")
            jobs.sort(reverse=True)
            ret = True
        except:
            logging.info(f"drucker jobs abfrage cups .backend.app.main -->  ERROR {ret}")
            ret = False
        return ret, jobs
    

    all_open_printer_jobs =[]
    have_jobs = False
    for pinter in printers:
        ret, jobs = load_print_jobs(pinter)
        if ret:
            jobs.sort(reverse=False)
            if len(jobs) > 0:
                have_jobs = True
            all_open_printer_jobs.append(jobs)

    if not have_jobs:
        flash(request, f"Kein Druckaufrtag offen", "flush_success")


    all_open_printer_jobs.sort(reverse=True)
    while [] in all_open_printer_jobs :
        all_open_printer_jobs.remove([])

    #logging.info(f"offen print jobs revers true --> {all_open_printer_jobs}")
    return templates.TemplateResponse("print_queue.html", {"request": request,
                                                           "all_open_printer_jobs":all_open_printer_jobs
                                                           })

####################
# projekte verwalten
@app.get("/projects", response_class=HTMLResponse)
def projects(request: Request,
              del_project: str = '',
              load_project: str = ''):
    
    path = 'static/_data/_projects'
    
    # Folder _projects erstellen falls nicht vorhanden
    if not os.path.exists(path):
        try: 
            os.mkdir(path)
            flash(request, "Ordner _projects erstellt", "flush_success")
        except OSError as error: 
            flash(request, f"Ordner _projects  konnte nicht erstellt werden - {error}", "flush_danger")

    # projekt löschen
    if not del_project == '':
        try:
            shutil.rmtree(os.path.join(path, del_project))            
            flash(request, f"Projekt << {del_project} >> gelöscht", "flush_success")
        except OSError as error: 
            flash(request, f"Projekt-Ordner konnte nicht gelöscht werden - {error}", "flush_danger")
    
    # projekt laden
    if not load_project == '':
        project_folder = os.path.join(path, load_project)
        suffixes = (".pkl", ".png",".pdf",".jpg")
        prefixes = ("Seite_")
        # Ordner _data und _print_jobs aufräumen
        dir = 'static/_print_jobs'
        for f in os.listdir(dir):
            os.remove(os.path.join(dir, f))
        dir = 'static/_data'
        for f in os.listdir(dir):
            if os.path.isfile(os.path.join(dir, f)):
                os.remove(os.path.join(dir, f))
        # Projekt-files laden
        for f in os.listdir(project_folder):
            if f.endswith(suffixes) and not f.startswith(prefixes):
                file_path_to_data = os.path.join('static/_data', f)
                file_path_from_store = os.path.join(project_folder, f)
                shutil.copyfile(file_path_from_store,file_path_to_data)
        
        for f in os.listdir(project_folder):
            if f.startswith(prefixes):
                file_path_to_data = os.path.join('static/_print_jobs', f)
                file_path_from_store = os.path.join(project_folder, f)
                shutil.copyfile(file_path_from_store,file_path_to_data)
                
        flash(request, f"Projekt <<-- {load_project} -->> geladen", "flush_success")
        
        return RedirectResponse("index")


    # Projektordner Listen
    projects = []
    for f in os.listdir(path):
        project_details = []
        project_print_para = os.path.join(path, f, '_print_para.pkl')
        ret, papier_format,papier_richtung,papier_rand, zentriert, \
        not_print, show_img_info, selcted_printer, spiegeln, \
        project_name,project_besch = load_print_para(project_print_para)
        project_details.append((project_name,project_besch))
        projects.append(project_details)

    projects.sort()
    if len(projects) < 1:
        flash(request, f"Noch keine Projekte gespeichert", "flush_success")


    ## paramer um an bildername an zuhängen um browserchache zu umgehen 
    cache_para = int(time.time())


    return templates.TemplateResponse("projects.html", {"request": request,
                                                        "cache_para": cache_para,
                                                    "projects": projects})

####################
# cups-pdf verwalten
@app.get("/cups_pdfs", response_class=HTMLResponse)
def projects(request: Request,
              del_pdf: str = '',
              open_pdf: str = '',
              download_pdf: str = ''):
    
    path = 'static/_data/_cups-pdf'

    ## paramer um an bildername an zuhängen um browserchache zu umgehen 
    cache_para = int(time.time())


    # pdf öffnen
    if not open_pdf =='':
        return templates.TemplateResponse("pdf_fullscreen.html", {"request": request,
                                                                "cache_para": cache_para,                                                                  
                                                                "open_pdf": open_pdf})
    # pdf herunterladen                                                            
    if not download_pdf =='':
        return FileResponse(f'{download_pdf}', media_type='application/octet-stream', filename='wrapper_cups.pdf')




    # pdf löschen
    if not del_pdf == '':
        try:
            os.remove(del_pdf)            
            flash(request, f"PDF << {del_pdf} >> gelöscht", "flush_success")
        except OSError as error: 
            flash(request, f"PDF {del_pdf} konnte nicht gelöscht werden - {error}", "flush_danger")
    
    # Projektordner Listen
    pdfs = []
    suffixes = (".pdf")        
    for f in os.listdir(path):
        if f.endswith(suffixes):
            pdfs.append(os.path.join(path, f))
        
    pdfs.sort(reverse=True)

    ## paramer um an bildername an zuhängen um browserchache zu umgehen 
    cache_para = int(time.time())


    return templates.TemplateResponse("cups_pdfs.html", {"request": request,
                                                        "cache_para": cache_para,
                                                        "pdfs": pdfs})


## dateien upload

@app.post("/uploadfile", response_class=HTMLResponse)
async def uploadfile(request: Request, file: UploadFile = File(...))-> None:
    
    # aufräumen image_files _data
    suffixes = (".png",".pdf",".jpg")
    dir = 'static/_data'
    for f in os.listdir(dir):
        if f.endswith(suffixes):
            os.remove(os.path.join(dir, f))

    
    #destination = Path(f'static/_data/ID_{datetime.datetime.today().strftime("%Y%m%d%H%M")}_{file.filename}')
    file_name, file_extension = os.path.splitext(file.filename)
    supported_files = ['.png','.jpg','.jpeg','.bmp','.tiff','.psd']

    if file_extension.lower() in supported_files:
        destination = Path(f'static/_data/ID_x{file_extension}')
        # logging.info(f"backend.app.main.uploadfile: --> inhalt destination: {destination}")
        try:
            with destination.open("wb") as buffer:
                shutil.copyfileobj(file.file, buffer)
        except:
            flash(request, "Upload {file} fehlgeschlagen", "flush_danger")

        finally:
            file.file.close()
        ### image zu png
        params = ['convert', destination, '-quality', '100', 'static/_data/ID_0.png']
        subprocess.check_call(params)
        flash(request, "Bild erfolgreich hochgeladen - Ziel Masse eingeben und auf Transformieren tippen - Hilfe zu den einzelnen Parametern unter Menüpunkt < ?! >", "flush_success")
    else:
        flash(request, f"Dateityp {file_extension} nicht unterstützt!", "flush_danger")
        #flash(request, f"Dateityp {file_extension} nicht unterstützt! - Unterstützte Formate: {supported_files}", "flush_danger")

    return RedirectResponse("index")

###################################################
## wrap ....
##

######
#  aufräumen wenn parameter editieren durch user am laufen
#
@app.get("/cleanup", response_class=HTMLResponse)
def clean_desk(request: Request,
                    U: int = 0,
                    u: int = 0,
                    m: int = 0,
                    dpi: int = 0
                    ):
    

    # ## alte parameter holen  ToDo func save_parameters noch umschreiben!!
    # #  
    # _, U_old, u_old, m_old, dpi_old = load_parameters()
    # if U > 0:
    #     u = u_old
    #     m = m_old
    #     dpi = dpi_old
    #     logging.info(f"U__proport.backend.app.main -->  {U}")
    # elif u > 0:
    #     U = U_old
    #     m = m_old
    #     dpi = dpi_old
    # elif  m > 0:
    #     u = u_old
    #     U = U_old
    #     dpi = dpi_old
    # elif dpi > 0:
    #     u = u_old
    #     m = m_old
    #     U = U_old

    # # geänderten wert speichern
    # #
    # try:
    #     save_parameters(U,u,m,dpi)
    #     logging.info(f"parameter in cleanup.backend.app.main -->  {U},{u},{m},{dpi}")

    # except NameError:
    #     logging.info(f"ERROR saveraram.cleanup.backend.app.main -->  {U}")


    # Ordner _data und _print_jobs aufräumen
    suffixes = ("_1.png","_flip.png",".pdf",".jpg")
    dir = 'static/_print_jobs'
    for f in os.listdir(dir):
        os.remove(os.path.join(dir, f))
    dir = 'static/_data'
    for f in os.listdir(dir):
        if os.path.isfile(os.path.join(dir, f)):
            if f.endswith(suffixes):
                os.remove(os.path.join(dir, f))

    return RedirectResponse("index")


###############
# Filegrösse der bilder bei "Übergrössen" reduzieren
#
def check_dpi(U_check, u_check, m_check, dpi=1200):

    ## Idee? - DPI anpassen zb.: default 300 | > 0.25m2 -> 150/144 | > 0.5 -> m2 72 
    #
    changed = False
    if U_check < u_check:
        U_check = u_check

    A = U_check*m_check/1000

    if A > 450:
        dpi_check = 72
    elif A > 150:
        dpi_check = 150
    elif A > 10:
        dpi_check = 300
    else:
        dpi_check = 600

    if dpi_check >= dpi:
        dpi_check = dpi
    else:
        changed = True

    return dpi_check, changed


###############
# parameter proportional ändern
#
@app.get("/proport", response_class=HTMLResponse)
def get_proportional(request: Request,
                    U: int = 0,
                    u: int = 0,
                    m: int = 0
                    ):
    
    ## alte parameter holen und proportional ... 
    #  
    _, U_old, u_old, m_old, _ = load_parameters()
    if U > 0:
        f = U/U_old
        u = round(u_old*f)
        m = round(m_old*f)
        logging.info(f"U__proport.backend.app.main -->  {U}")
    elif u > 0:
        f = u/u_old
        U = round(U_old*f)
        m = round(m_old*f)
    elif  m > 0:
        f = m/m_old
        U = round(U_old*f)
        u = round(u_old*f)

    ##############
    # check dpi
    dpi, changed = check_dpi(U, u, m)
    if changed:
        flash(request, f"Auf proportional ->>>>>>>  Grund der Zielgrösse wurde der DPI-Wert auf {dpi} geändert!", "flush_danger")

    ## save
    #
    try:
        save_parameters(U,u,m,dpi)

    except NameError:
        logging.info(f"ERROR saveraram.proport.backend.app.main -->  {U}, {u}, {m},{dpi}")

    return RedirectResponse("index")


@app.get("/trim", response_class=HTMLResponse)
def trim(request: Request
                    ):
    ## bild trimmen 
    shutil.copyfile('static/_data/ID_0.png', 'static/_data/ID_0_pre_trim.png') # kopieren für evtl. download / restore
    params = ['convert',
            'static/_data/ID_0.png',
            '-fuzz',
            '7%',
            '-trim',
            '+repage',
            'static/_data/ID_0.png']
    try:
        subprocess.check_call(params)
        flash(request, "Bild erfolgreich zugeschnitten", "flush_success")
    except OSError as error: 
        flash(request, f"Zuschnitt des Bild missglückt!  - {error}", "flush_danger")

    return RedirectResponse("index")




#####################################
###         Transform Kegel       ###
###################################3#
@app.get("/kegelstumpf-abwicklung", response_class=HTMLResponse)
def get_alpha(request: Request, 
                U: int = 276,
                u: int = 230,
                m: int = 50,
                dpi: int =300,
                src_file: str = "static/_data/ID_0.png",
                spiegeln: bool = True,
                img_reset: bool = False):   
    
    # Masse aus originalbild übernehmen und in website übernehmen
    if img_reset:
        img_source_path = 'static/_data/ID_0.png'
        if not os.path.isfile(img_source_path):
                return RedirectResponse("index")
        flash(request,"Werte auf Masse des Originalbild zurückgesetzt --->", "flush_success")
        img_src = Image.open(img_source_path)
        dpi = round(img_src.info['dpi'][0])
        w_src = int(img_src.width)
        h_src = int(img_src.height)
        U, m = pix2mm(w_src, h_src, dpi, dpi)
        u = U

    ##############
    # Gewählte Bildgrösse überwachen
    dpi, changed = check_dpi(U, u, m, dpi)
    if changed:
        flash(request, f"Auf Grund der Bildgrösse wurde der DPI-Wert auf {dpi} gesetzt!", "flush_danger")


    save_parameters(U,u,m,dpi)

    flipped = False
    ## bild drehen wenn nach oben gebogen werden soll
    if u > U:
        tempU = U
        U = u
        u = tempU
        flipped = True
    
    ### alpha kegelstumpf berechnen
    pi = 3.141592653589793
    inch2mm = 25.4
    Upixel = int(U*dpi/inch2mm)
    mPixel = int(m*dpi/inch2mm)

    ### Bild scalieren
    str_dim = str(f'{Upixel}x{mPixel}!')
    params = ['convert',
                src_file,
                '-resize',
                str_dim,
                'static/_data/ID_1.png']
    subprocess.check_call(params)
    
    ## bild drehen wenn nach oben gebogen werden soll
    shutil.copyfile('static/_data/ID_1.png', 'static/_data/ID_1_pre_flip.png') # der einfachheit wegen immer kopieren für evtl. download
    if flipped:
        params = ['convert',
                'static/_data/ID_1.png',
                '-flip',
                'static/_data/ID_1.png']
        subprocess.check_call(params)


    ### Bild biegen
    ## sonderfall U == u -> heisst linien kreuzen sich nie = ERROR
    if U == u:
        u = u - 0.0001

    R = U/(2*pi)
    r = u/(2*pi)
    alpha = 0

    M = m*R/(R-r)
    M_pixel = int(M*dpi/inch2mm)
    m_pixel = int(M_pixel - mPixel)
    #alpha = 360*R/M
    alpha = (180*U)/(pi*M)


    # logging.info(f"R__backend.app.main -->  {R}")
    # logging.info(f"r__backend.app.main -->  {r}")
    # logging.info(f"M__backend.app.main -->  {M}")
    # logging.info(f"M_pixel__backend.app.main -->  {M_pixel}")
    # logging.info(f"m_pixel__backend.app.main -->  {m_pixel}")
    # logging.info(f"alpha__backend.app.main -->  {alpha}")

    params = ['convert', 'static/_data/ID_1.png', '-virtual-pixel', 'White','-flatten', '-distort','Arc', f'{alpha} 0 {M_pixel} {m_pixel}', '-quality', '100', '-units', 'PixelsPerInch',  '-density', str(dpi), 'static/_data/ID_2.jpg']
    subprocess.check_call(params)
    flash(request, f"Erfolgreich tranformiert - Das Bild steht unter den Menupunkten < Drucken > und < Downloads > bereit.", "flush_success")

    ## bild zurück drehen wenn nach oben gebogen wurde
    if flipped:
        params = ['convert',
                'static/_data/ID_2.jpg',
                '-flip',
                'static/_data/ID_2.jpg']
        subprocess.check_call(params)

    ### spiegeln -fop
    if spiegeln:
        #logging.info(f"spiegeln?_index.app.main -->  {spiegeln} ---> gespiegelt")
        params = ['convert', 'static/_data/ID_2.jpg','-flop', '-quality', '100', 'static/_data/ID_2s.jpg']
        subprocess.check_call(params)
    else:
        shutil.copyfile('static/_data/ID_2.jpg', 'static/_data/ID_2s.jpg') # der einfachheit wegen immer kopieren für evtl. download

    ### res zu pdf
    params = ['convert', 'static/_data/ID_2s.jpg','-quality', '100', 'static/_data/ID_2s.pdf']
    subprocess.check_call(params)
    params = ['convert', 'static/_data/ID_2.jpg','-quality', '100', 'static/_data/ID_2.pdf']
    subprocess.check_call(params)


    return RedirectResponse("index")

######## downloads
#
@app.get("/downloads", response_class=HTMLResponse)
def downloads(request: Request):
    ## paramer um an bildername anzuhängen um browserchache zu umgehen 
    cache_para = int(time.time())
    dir = 'static/_print_jobs'
    suffixes = (".pdf", ".PDF")
    pdfs =[]
    for r, d, f in os.walk(dir):
        for file in f:
            # NULL bit files hier löschen
            b = os.stat(os.path.join(r, file)).st_size
            if b == 0:
                print('löschen - ist null')
            if file.endswith(suffixes):
                pdfs.append(os.path.join(r, file))
    pdfs.sort()

    return templates.TemplateResponse("download.html", {"request": request,
                                                        "cache_para": cache_para,
                                                        "pdfs": pdfs})

@app.get("/download-druckseite")
def download_file(request: Request,
                  druckseite: str = None):
    filename = druckseite
    return FileResponse(druckseite, media_type='application/octet-stream', filename=filename)

@app.get("/download-conferted")
def download_file_conferted():
    return FileResponse('static/_data/ID_0.png', media_type='application/octet-stream', filename='ID_0.png')

@app.get("/download-resized")
def download_file_resized():
    return FileResponse('static/_data/ID_1.png', media_type='application/octet-stream', filename='ID_1.png')

@app.get("/download-wraped")
def download_file_wraped():
    return FileResponse('static/_data/ID_2.jpg', media_type='application/octet-stream', filename='ID_2.jpg')

@app.get("/download-wraped_s")
def download_file_wraped_s():
    return FileResponse('static/_data/ID_2s.jpg', media_type='application/octet-stream', filename='ID_2s.jpg')

@app.get("/download-pdf")
def download_file_pdf():
    return FileResponse('static/_data/ID_2s.pdf', media_type='application/octet-stream', filename='ID_2s.pdf')

@app.get("/download-wraped-pdf")
def download_file_wraped_pdf():
    return FileResponse('static/_data/ID_2.pdf', media_type='application/octet-stream', filename='ID_2.pdf')


###################
# drucken

@app.get("/not_to_print", response_class=HTMLResponse)
def print_center(request: Request,
                    not_to_print: int = None
                    ):
    ## abbruch wenn not_to_print leer none
    if not_to_print == None:
        return RedirectResponse("print")
    
    # gespeicherte druckparameter lesen
    ret, papier_format,papier_richtung,papier_rand, zentriert, \
    not_print,show_img_info,selcted_printer,spiegeln,project_name,project_besch = load_print_para()

    if not_print == None:
        not_print = []

    if not_to_print not in not_print:
        not_print.append(not_to_print)
    else:
        not_print.remove(not_to_print)

    #################
    # parameter sichern
    save_print_para(papier_format=papier_format,
                        papier_richtung=papier_richtung,
                        papier_rand=papier_rand,
                        zentriert=zentriert,
                        not_print=not_print,
                        show_img_info=show_img_info,
                        selcted_printer=selcted_printer,
                        spiegeln=spiegeln,
                        project_name=project_name,
                        project_besch=project_besch)

    
    return RedirectResponse("print")



@app.get("/print", response_class=HTMLResponse)
def print_center(request: Request,
                    papier_format: str = None,
                    papier_richtung: str = None,
                    papier_rand: int = None,
                    zentriert: int = -1,
                    not_print: typing.Annotated[list[int] | None, Query()] = [],
                    show_img_info: str = None,
                    print_jobs_erstellen: bool = False,
                    selcted_printer: str = None,
                    spiegeln: str = 'off',
                    project_name: str = '',
                    project_besch: str =''):
    
    #https://fastapi.tiangolo.com/tutorial/query-params-str-validations/#__tabbed_15_1
    # params = str(request.query_params)
    # logging.info(f"print: --> request.params = {params}")
    # logging.info(f"print: --> spiegeln = {spiegeln}")

    #################
    # parameter sichern
    if papier_format != None:
        save_print_para(papier_format=papier_format,
                        papier_richtung=papier_richtung,
                        papier_rand=papier_rand,
                        zentriert=zentriert,
                        not_print=not_print,
                        show_img_info=show_img_info,
                        selcted_printer=selcted_printer,
                        spiegeln=spiegeln,
                        project_name=project_name,
                        project_besch=project_besch)

    # cups druckerliste abfragen
    ret, printers = load_printer_list()
    if not ret:
        printers =['Druckserver CUPS nicht erreichbar']
        flash(request, "Der Druckserver 'cups' ist nicht erreichbar!", "flush_danger")


    ##########################
    # aufräumen
    # alte files löschen in _print_preview
    dir = 'static/_print_preview'
    for f in os.listdir(dir):
        os.remove(os.path.join(dir, f))


    # gespeicherte druckparameter lesen
    ret, papier_format,papier_richtung,papier_rand, zentriert, \
    not_print,show_img_info,selcted_printer,spiegeln,project_name,project_besch = load_print_para()

    #############################
    #
    # zu drucken des bild wählen - gespiegel oder nicht gespiegel
    #
    if spiegeln == 'on':
        img_source_path = 'static/_data/ID_2s.jpg'
    else:
        img_source_path = 'static/_data/ID_2.jpg'


    ##############################
    #
    # infos sammeln
    # 
    # bildgrösse auflösung aus src_bild bestimmen
    if not os.path.isfile(img_source_path):
            return RedirectResponse("index")
    img_src = Image.open(img_source_path)
    dpi_src_w, dpi_src_h = round(img_src.info['dpi'][0]), round(img_src.info['dpi'][1])
    w_src = int(img_src.width)
    h_src = int(img_src.height)
    wmm_src, hmm_src = pix2mm(w_src, h_src, dpi_src_w, dpi_src_h)
    # logging.info(f"src_breit_backend.app.main -->  {w_src} pix")
    # logging.info(f"src_hoch_backend.app.main -->  {h_src} pix")
    # logging.info(f"src_dpi_backend.app.main -->  {dpi_src} dpi")
    # logging.info(f"wmm_src_backend.app.main -->  {wmm_src} mm")
    # logging.info(f"hmm_src_backend.app.main -->  {hmm_src} mm")


    # DIN A - formate
    DIN_From = {"A0": [840, 1189], "A1": [594, 840], "A2": [420, 594], "A3": [297,420], "A4": [210, 297], "A5":	[148, 210], "A6": [105, 148]}

    w_blatt = DIN_From[papier_format][0] - papier_rand*2
    h_blatt = DIN_From[papier_format][1] - papier_rand*2
    w_blatt_mit_rand = DIN_From[papier_format][0]
    h_blatt_mit_rand = DIN_From[papier_format][1]
    if papier_richtung == 'quer':
        h_blatt, w_blatt = w_blatt, h_blatt
        h_blatt_mit_rand, w_blatt_mit_rand = w_blatt_mit_rand, h_blatt_mit_rand
    # logging.info(f"w_blatt_backend.app.main -->  {w_blatt} mm")
    # logging.info(f"h_blatt_backend.app.main -->  {h_blatt} mm")

    # Erste grobe verteilung auf blätter - seitenrand - anzahl bätter wie papierrichtung in auwahl durch user gesetzt
    f_wimg_wblatt = wmm_src / w_blatt
    f_himg_hblat = hmm_src / h_blatt
    # aufrunden 
    anzahl_blatt_w = math.ceil(f_wimg_wblatt)
    anzahl_blatt_h = math.ceil(f_himg_hblat)
    anzahl_blatt_T = anzahl_blatt_w * anzahl_blatt_h
    w_mm_blaetter = w_blatt * anzahl_blatt_w
    h_mm_blaetter = h_blatt * anzahl_blatt_h
    # logging.info(f"f_wimg_wblatt.backend.app.main -->  {f_wimg_wblatt} = {anzahl_blatt_w} breit")
    # logging.info(f"f_himg_hblat.backend.app.main -->  {f_himg_hblat} = {anzahl_blatt_h} hoch")

    ## blätter als linien einzeichnen
    # 1. bild clonen
    img_markiert = img_src.copy()
    
    # 2. Leeres Blatt grösse wie blätter mal blätter - img_source einfügen
    # 
    w, h = mm2pix(w_mm_blaetter,h_mm_blaetter, dpi_src_h)
    im = Image.new(mode = "RGB", size = (w, h),color = (255, 255, 255))
    # im.save('static/_data/hintergrund.jpg', dpi=(dpi_src_w, dpi_src_h))
    # 1. src bild auf hintergund legen
    # logging.info(f"zentriert.backend.app.main -->  {zentriert}")
    if zentriert == 0:
        x,y = int(w/2-w_src/2), int(h/2-h_src/2)
        im.paste(img_markiert, (x,y))
    elif zentriert == -2: #horizontal
        x,y = 0, int(h/2-h_src/2)
        im.paste(img_markiert, (x,y))
    elif zentriert == -3: # vertikal
        x,y = int(w/2-w_src/2), 0
        im.paste(img_markiert, (x,y))
    elif zentriert > 0:
        i =0
    else:
        im.paste(img_markiert)

    # grosses bild 'im' clonen in im_for_Print_crop -> um ausschnitte für printjobs daues zu schneiden
    im_for_print_crops = im.copy()
    # blätter umranden und koordinaten merken
    # Stift auf ganzes Blatt/HINTERGRUNDBLATT
    draw_img1 = ImageDraw.Draw(im)  

    # funktion Text einfügen
    def past_text(img=None, pos=(0,0), text=" ",text_groesse=10, nr = 0):
        myfont = ImageFont.truetype(r'static/fonts/POE Sans Pro Heavy.ttf', int(text_groesse))
        mergine = dpi_src_w/6
        bbox = img.textbbox((pos[0]+mergine,pos[1]+mergine), ' ' + text+str(nr) +' ', font=myfont)
        img.rectangle(bbox, fill="red",outline ="yellow")
        img.text((pos[0]+mergine,pos[1]+mergine), text+str(nr), fill=(0, 0, 0), font=myfont)

    # printjob an cups
    def send_to_cups(file_path='static/_print_jobs/Seite_0.pdf', region=None, media='A4',ausrichtung="quer"):        
        # unterscheidung hoch quer hier unnütz und eher schlechte idee - ohne effekt - cups rottiert zurück was auch je drucker sehr sinnvoll ist
        # pdfinfo _print_jobs/Seite_R_0.pdf  - https://wiki.debian.org/CUPSPdfToPdf
        # if ausrichtung == 'quer':
        #     params = ['lp', '-h', 'cups', '-d', f'{selcted_printer}', '-o', 'landscape','-o',f'media={media}', f'{file_path}']
        # else:
        params = ['lp', '-h', 'cups', '-d', f'{selcted_printer}','-o',f'media={media}', f'{file_path}']
        try:
            # druck ID herausfiltern
            ret = str(subprocess.check_output(params))
            start = ret.index(f'{selcted_printer}-') + len(f'{selcted_printer}-')
            id_start = ret[start:]
            id = id_start[:id_start.index(' ')]
            # vorschaubild für druckauftrag erstellen und speichern
            size = (300, 300)
            region.thumbnail(size)
            region.save(f'static/_open_print_jobs/{id}.png')
        except:
            flash(request, "Übermittlung Druckauftrag an {selcted_printer} fehlgeschlagen", "flush_danger")
            logging.info(f"pdfs-an-drucker-senden.backend.app.main -->  ERROR")

    # iterieren über Blätter
    rechteck = []   #(w_o,h_0),(w_u,h_u)
    rechtecke = []
    bild_nr = 0
    dicke, _ = mm2pix(1,1,dpi_src_w)

    # alte files löschen in _print_jobs falls neuer job starten soll
    if print_jobs_erstellen:
        dir = 'static/_print_jobs'
        for f in os.listdir(dir):
            os.remove(os.path.join(dir, f))

    for hoehe in range(0,anzahl_blatt_h):
        for weite in range(0,anzahl_blatt_w):
            # links oben zu rechts unten
            #x1, y1, x2, y2  = (0,0,180,267)
            x1,y1 = mm2pix(w_blatt*weite, h_blatt*hoehe, dpi_src_h)
            x2,y2 = mm2pix(w_blatt*(weite+1), h_blatt*(hoehe+1), dpi_src_h)

            rechteck  = [(x1,y1),(x2,y2)]
            rechtecke.append(rechteck)

            if not_print:
                dicke_red = round(((x2-x1)+(y2-y1)/2)*0.03)
                if bild_nr in not_print:
                    draw_img1.line(((x1,y1),(x2,y2)), fill="red", width=dicke_red)
                    draw_img1.line(((x1,y2),(x2,y1)), fill="red", width=dicke_red)
            
            # Wenn Druck Job ausgeführt werden soll Bilder erstellen und senden
            if print_jobs_erstellen:
                # Seite ausschneiden region
                region = im_for_print_crops.crop((x1,y1,x2,y2))
                
                # Rand weiss hinzufügen blatt_weiss = hintergrund genau papierformat zb A4:
                # w_blatt_mit_rand_in_pix, h_blatt_mit_rand_in_pix = mm2pix(w_blatt+2*papier_rand,h_blatt+2*papier_rand,dpi_src_h)
                w_blatt_mit_rand_in_pix, h_blatt_mit_rand_in_pix = mm2pix(w_blatt_mit_rand,h_blatt_mit_rand,dpi_src_h)
                rand_in_pix, _ = mm2pix(papier_rand,papier_rand,dpi_src_w)
                # w_mit_rand = w+2*rand_in_pix
                # h_mit_rand = h+2*rand_in_pix
                blatt_mit_rand = Image.new(mode = "RGB", size = (w_blatt_mit_rand_in_pix, h_blatt_mit_rand_in_pix),color = (255, 255, 255))
                # regio auf blatt mit rand legen
                blatt_mit_rand.paste(region, (rand_in_pix,rand_in_pix))

                # entscheid ob verworfen werden soll oder doch drucken -> crop noch nach unten scheiben
                def create_pdfs(bild_nr_):
                    # ToDo - pdf mit oder ohne rand????
                    # _  region.save(f"static/_print_jobs/Seite_{bild_nr_}.pdf",dpi=(dpi_src_w,dpi_src_h))
                    blatt_mit_rand.save(f"static/_print_jobs/Seite_R_{bild_nr_}.pdf",dpi=(dpi_src_w,dpi_src_h))

                # ToDo - druckoption für development
                m_o_o_R = 'R_' # '' oder 'R_'  = PDF mit oder ohne Rand drucken??????? 

                if not_print:
                    if not bild_nr in not_print:
                        create_pdfs(bild_nr)
                        if not selcted_printer == 'nurPDFerstellen_254242859125925':
                            send_to_cups(f"static/_print_jobs/Seite_{m_o_o_R}{bild_nr}.pdf", blatt_mit_rand, papier_format, papier_richtung)
                else:
                    create_pdfs(bild_nr)
                    if not selcted_printer == 'nurPDFerstellen_254242859125925':
                        send_to_cups(f"static/_print_jobs/Seite_{m_o_o_R}{bild_nr}.pdf", blatt_mit_rand, papier_format, papier_richtung)

            
            draw_img1.line(((x1,y1),(x2,y1),(x2,y2),(x1,y2),(x1,y1)), fill="lightgreen", width=dicke)
            draw_img1.rectangle(rechteck, outline ="red")

            bild_nr += 1
            past_text(draw_img1,(x1,y1)," S ",nr=bild_nr, text_groesse=dicke*4)


    if print_jobs_erstellen:
        if selcted_printer == 'nurPDFerstellen_254242859125925':
            flash(request, f"PDF erstellt - Unter  Menüpunkt: < Downloads > abrufbar. - TIPP : Alle Einstellungen und Bilder hier als Projekt speichern.", "flush_success")
        elif selcted_printer == 'CUPS-PDF':
            flash(request, f"An CUPS-PDF gesendet - PDF unter Menüpunkt: < CUPS-PDF > abrufbar. - TIPP : Alle Einstellungen und Bilder hier als Projekt speichern.", "flush_success")
        else:
            flash(request, f"An Drucker {selcted_printer} gesendet - Verwaltung der Warteschlange unter Menüpunkt: < Druckaufträge > - TIPP : Alle Einstellungen und Bilder hier als Projekt speichern.", "flush_success")



    # Umrandung  Hintergrund_Blatt
    shape = [(0, 0), (w-1, h-1)]
    draw_img1.rectangle(shape, outline ="red")
     
    # draw(img_markiert)
    im.save('static/_print_preview/markiert.jpg', dpi=(dpi_src_w, dpi_src_h))

    jpg='static/_print_preview/markiert.jpg'

    return templates.TemplateResponse("print.html", {"request": request,
                                                     "jpg": jpg,
                                                     "papier_format":papier_format,
                                                     "papier_richtung": papier_richtung,
                                                     "papier_rand": papier_rand,
                                                     "dpi_src_w": dpi_src_w,
                                                     "dpi_src_h": dpi_src_h,
                                                     "w_src": w_src,
                                                     "h_src": h_src,
                                                     "wmm_src": wmm_src,
                                                     "hmm_src":hmm_src,
                                                     "w_blatt": w_blatt,
                                                     "h_blatt":h_blatt,
                                                     "anzahl_blatt_w": anzahl_blatt_w,
                                                     "anzahl_blatt_h": anzahl_blatt_h,
                                                     "anzahl_blatt_T": anzahl_blatt_T,
                                                     "w_mm_blaetter": w_mm_blaetter,
                                                     "h_mm_blaetter": h_mm_blaetter,
                                                     "zentriert": zentriert,
                                                     "not_print": not_print,
                                                     "show_img_info": show_img_info,
                                                     "selcted_printer": selcted_printer,
                                                     "printers": printers,
                                                     "spiegeln": spiegeln,
                                                     "project_name": project_name,
                                                     "project_besch":project_besch})
